using System;
using System.Text;

namespace Printy.Library
{
  /// <summary>
  /// Diese Klasse speichert einen Bitvektor mit Indizes des Intervalls [0;64[.
  /// Sie bietet bequeme Operator-berladungen an.
  /// </summary>
  public struct BitSet
  {
    /// <summary>
    /// ToString(): gibt die String-Reprsentation eines gesetzten Index zurck.
    /// </summary>
    public delegate string StringMapper(int index);

    /// <summary>
    /// Gibt den Index als String zurck. Dies ist das Standardverfahren.
    /// </summary>
    /// <param name="index"></param>
    /// <returns></returns>
    public static string DefaultMapper(int index)
    {
      return index.ToString();
    }

    /// <summary>
    /// berprft den Index und wirft bei Verletzung der Grenzen eine IndexOutOfRangeException.
    /// </summary>
    /// <param name="index">zu prfender Index</param>
    static void CheckIndex(int index)
    {
      if ((index < 0) || (index >= 64))
      {
        throw new IndexOutOfRangeException(index.ToString());
      }
    }

    /// <summary>
    /// Addition
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="index">zu setzender Index</param>
    /// <returns>BitSet mit allen Indizes aus a sowie index auf true</returns>
    public static BitSet operator + (BitSet a, int index)
    {
      CheckIndex(index);
      return new BitSet(a.bitmask | (1UL << index));
    }
    
    /// <summary>
    /// Addition
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>BitSet mit allen Indizes aus a sowie allen Indizes aus b auf true</returns>
    public static BitSet operator + (BitSet a, BitSet b)
    {
      return new BitSet(a.bitmask | b.bitmask);
    }

    /// <summary>
    /// Subtraktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="index">zu lschender Index</param>
    /// <returns>BitSet mit allen Indizes aus a auf true sowie index auf false</returns>
    public static BitSet operator - (BitSet a, int index)
    {
      CheckIndex(index);
      return new BitSet(a.bitmask & ~(1UL << index));
    }

    /// <summary>
    /// Subtraktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet mit zu lschenden Indizes</param>
    /// <returns>BitSet mit allen Indizes aus a auf true sowie allen Indizes aus b auf false </returns>
    public static BitSet operator - (BitSet a, BitSet b)
    {
      return new BitSet(a.bitmask & ~b.bitmask);
    }

    /// <summary>
    /// Konjuktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>BitSet mit allen Indizes aus beiden BitSet Objekten</returns>
    public static BitSet operator | (BitSet a, BitSet b)
    {
      return new BitSet(a.bitmask | b.bitmask);
    }

    /// <summary>
    /// Konjunktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">Wahrheitswert</param>
    /// <returns>falls b true ist, BitSet mit allen mglichen Indizes, sonst mit allen Indizes aus a</returns>
    public static BitSet operator | (BitSet a, bool b)
    {
      return new BitSet(b ? ulong.MaxValue : a.bitmask);
    }

    /// <summary>
    /// Disjunktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>BitSet mit allen Indizes, die sowohl in a als auch in b true sind.</returns>
    public static BitSet operator & (BitSet a, BitSet b)
    {
      return new BitSet(a.bitmask & b.bitmask);
    }

    /// <summary>
    /// Disjunktion
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">Wahrheitswert</param>
    /// <returns>falls b true ist, BitSet mit allen mglichen Indizes, ansonsten leeres BitSet Objekt</returns>
    public static BitSet operator & (BitSet a, bool b)
    {
      return new BitSet(b ? a.bitmask : 0);
    }

    /// <summary>
    /// XOR
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="index">zu invertierender Index</param>
    /// <returns>BitSet mit allen Indizes, die entweder in a gesetzt sind oder index entsprechen</returns>
    public static BitSet operator ^ (BitSet a, int index)
    {
      CheckIndex(index);
      return new BitSet(a.bitmask ^ (1UL << index));
    }

    /// <summary>
    /// XOR
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>BitSet mit allen Indizes, die entweder in a oder in b auf true sind</returns>
    public static BitSet operator ^ (BitSet a, BitSet b)
    {
      return new BitSet(a.bitmask ^ b.bitmask);
    }

    /// <summary>
    /// Komplement
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <returns>BitSet mit allen Indizes auf true, die in a auf false sind</returns>
    public static BitSet operator ~ (BitSet a)
    {
      return new BitSet(~a.bitmask);
    }

    /// <summary>
    /// Test auf Gleichheit
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>true, falls a und b dieselben Bits aufweisen</returns>
    public static bool operator == (BitSet a, BitSet b)
    {
      return a.bitmask == b.bitmask;
    }

    /// <summary>
    /// Test auf Ungleichheit
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <param name="b">BitSet</param>
    /// <returns>true, falls a und b unterschiedliche Bits aufweisen</returns>
    public static bool operator != (BitSet a, BitSet b)
    {
      return a.bitmask != b.bitmask;
    }

    /// <summary>
    /// Auswertungsoperator
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <returns>true, falls a irgendwelche Bits enthlt</returns>
    public static bool operator true(BitSet a)
    {
      return a.bitmask != 0;
    }

    /// <summary>
    /// Auswertungsoperator
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <returns>true, falls a keine Bits enthlt</returns>
    public static bool operator false(BitSet a)
    {
      return a.bitmask == 0;
    }

    /// <summary>
    /// Konvertierungsoperator
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <returns>true, falls a irgendwelche Bits enthlt</returns>
    public static explicit operator bool(BitSet a)
    {
      return a.bitmask != 0;
    }

    /// <summary>
    /// Konvertierungsoperator
    /// </summary>
    /// <param name="a">BitSet</param>
    /// <returns>Die Bitmaske</returns>
    public static explicit operator ulong(BitSet a)
    {
      return a.bitmask;
    }

    /// <summary>
    /// Bitmaske der gesetzten Bits.
    /// </summary>
    public ulong bitmask;

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="bitmask">Die Bitmaske fr das neue BitSet Objekt</param>
    public BitSet(ulong bitmask)
    {
      this.bitmask = bitmask;
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="indices">zu setzende Indizies</param>
    public BitSet(params int[] indices)
    {
      bitmask = 0UL;
      for (int i = 0; i < indices.Length; i++)
      {
        int index = indices[i];
        CheckIndex(index);
        bitmask |= 1UL << index;
      }
    }

    /// <summary>
    /// Vergleicht dieses BitSet mit einem Objekt.
    /// Gleichheit ist dann gegeben, falls das Objekt ein BitSet ist und
    /// dieselben Bits gesetzt hat.
    /// </summary>
    /// <param name="obj">Vergleichsobjekt</param>
    /// <returns></returns>
    public override bool Equals(object obj)
    {
      return (obj is BitSet) && (this == (BitSet)obj);
    }

    /// <summary>
    /// Gibt einen Hashcode zurck, der aus der Bitmaske generiert wird.
    /// </summary>
    /// <returns>XOR aus linker und rechter 32-Bit-Hflte der Bitmaske</returns>
    public override int GetHashCode()
    {
      return (int)((bitmask >> 32) ^ bitmask);
    }

    /// <summary>
    /// Gibt einen String mit einer Auflistung der gesetzten Bits zurck.
    /// </summary>
    /// <returns>Einen String mit den gesetzten Indizes</returns>
    public override string ToString()
    {
      return ToString(new StringMapper(DefaultMapper));
    }

    /// <summary>
    /// Gibt einen String mit einer Auflistung der gesetzten Bits zurck.
    /// Die Reprsentation des einzelnen Index wird vom StringMapper delegate
    /// bernommen.
    /// </summary>
    /// <returns>Einen String mit den gesetzten Indizes</returns>
    public string ToString(StringMapper mapper)
    {
      StringBuilder result = new StringBuilder();
      string sep = "";
      result.Append("[");
      for (int index = 0; index < 64; index++)
      {
        ulong mask = 1UL << index;
        if ((bitmask & mask) == mask)
        {
          result.Append(sep);
          result.Append(mapper(index));
          sep = ",";
        }
      }
      result.Append("]");
      return result.ToString();
    }

    /// <summary>
    /// Indexer, mit dem ein Bit ein- oder ausgeschalten oder gelesen werden kann.
    /// </summary>
    public bool this[int index]
    {
      get
      {
        CheckIndex(index);
        ulong mask = 1UL << index;
        return (bitmask & mask) == mask;
      }
	
      set
      {
        CheckIndex(index);
        ulong mask = 1UL << index;
        bitmask = value ? (bitmask | mask) : (bitmask & ~mask);
      }
    }
  }
}
