
using System;
using System.IO;
using System.Collections;

namespace Printy.Library {

public class Token {
	public int kind;    // token kind
	public int pos;     // token position in the source text (starting at 0)
	public int col;     // token column (starting at 0)
	public int line;    // token line (starting at 1)
	public string val;  // token value
	public Token next;  // ML 2005-03-11 Tokens are kept in linked list
	
	internal Token comment;  // PG 2005-06-27 pointer to a comment token
  internal int lines;      // PG 2005-06-27 number of lines of a comment token
}

public class Buffer {
	public const char EOF = (char)256;
	const int MAX_BUFFER_LENGTH = 64 * 1024; // 64KB
	byte[] buf;         // input buffer
	int bufStart;       // position of first byte in buffer relative to input stream
	int bufLen;         // length of buffer
	int fileLen;        // length of input stream
	int pos;            // current position in buffer
	Stream stream;      // input stream (seekable)
	bool isUserStream;  // was the stream opened by the user?
	
	public Buffer (Stream s, bool isUserStream) {
		stream = s; this.isUserStream = isUserStream;
		fileLen = bufLen = (int) s.Length;
		if (stream.CanSeek && bufLen > MAX_BUFFER_LENGTH) bufLen = MAX_BUFFER_LENGTH;
		buf = new byte[bufLen];
		bufStart = Int32.MaxValue; // nothing in the buffer so far
		Pos = 0; // setup buffer to position 0 (start)
		if (bufLen == fileLen) Close();
	}
	
	~Buffer() { Close(); }
	
	void Close() {
		if (!isUserStream && stream != null) {
			stream.Close();
			stream = null;
		}
	}
	
	public int Read () {
		if (pos < bufLen) {
			return buf[pos++];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos++];
		} else {
			return EOF;
		}
	}

	public int Peek () {
		if (pos < bufLen) {
			return buf[pos];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos];
		} else {
			return EOF;
		}
	}
	
	public string GetString (int beg, int end) {
		int len = end - beg;
		char[] buf = new char[len];
		int oldPos = Pos;
		Pos = beg;
		for (int i = 0; i < len; i++) buf[i] = (char) Read();
		Pos = oldPos;
		return new String(buf);
	}

	public int Pos {
		get { return pos + bufStart; }
		set {
			if (value < 0) value = 0; 
			else if (value > fileLen) value = fileLen;
			if (value >= bufStart && value < bufStart + bufLen) { // already in buffer
				pos = value - bufStart;
			} else if (stream != null) { // must be swapped in
				stream.Seek(value, SeekOrigin.Begin);
				bufLen = stream.Read(buf, 0, buf.Length);
				bufStart = value; pos = 0;
			} else {
				pos = fileLen - bufStart; // make Pos return fileLen
			}
		}
	}
}

public class Scanner {
	const char EOL = '\n';
	const int eofSym = 0; /* pdt */
	const int charSetSize = 256;
	const int maxT = 24;
	const int noSym = 24;
	short[] start = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  2,  0, 16,  0,  0,  5, 31, 22,  0,  0,  0,  0, 32, 17,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 27, 20, 28,  0,
	  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
	  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 23,  0, 24,  0,  0,
	  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
	  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 25, 21, 26,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  -1};


	public Buffer buffer; // scanner buffer
	
	Token t;          // current token
	char ch;          // current input character
	int pos;          // column number of current character
	int line;         // line number of current character
	int lineStart;    // start position of current line
	int oldEols;      // EOLs that appeared in a comment;
	BitArray ignore;  // set of characters to be ignored by the scanner

	Token tokens;     // list of tokens already peeked (first token is a dummy)
	Token pt;         // current peek token
	
	char[] tval = new char[128]; // text of current token
	int tlen;         // length of current token
	
	/// <summary>
	/// PG 2005-07-04
	/// Gibt die aktuelle Zeile des Scanners zurück;
	/// Dies wird gebraucht, um mehrzeilige Kommentare effizienter zu erkennen 
	/// </summary>
  internal int Line { get { return line; }}
	
	public Scanner (string fileName) {
		try {
			Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
			buffer = new Buffer(stream, false);
			Init();
		} catch (IOException) {
			Console.WriteLine("--- Cannot open file {0}", fileName);
			System.Environment.Exit(1);
		}
	}
	
	public Scanner (Stream s) {
		buffer = new Buffer(s, true);
		Init();
	}
	
	void Init() {
		pos = -1; line = 1; lineStart = 0;
		oldEols = 0;
		NextCh();
		ignore = new BitArray(charSetSize+1);
		ignore[' '] = true;  // blanks are always white space
		ignore[9] = true; ignore[10] = true; ignore[13] = true; 
		pt = tokens = new Token();  // first token is a dummy
	}
	
	void NextCh() {
		if (oldEols > 0) { ch = EOL; oldEols--; } 
		else {
			ch = (char)buffer.Read(); pos++;
			// replace isolated '\r' by '\n' in order to make
			// eol handling uniform across Windows, Unix and Mac
			if (ch == '\r' && buffer.Peek() != '\n') ch = EOL;
			if (ch == EOL) { line++; lineStart = pos + 1; }
		}

	}

	void AddCh() {
		if (tlen >= tval.Length) {
			char[] newBuf = new char[2 * tval.Length];
			Array.Copy(tval, 0, newBuf, 0, tval.Length);
			tval = newBuf;
		}
		tval[tlen++] = ch;
		NextCh();
	}




	void CheckLiteral() {
		switch (t.val) {
			case "COMPILER": t.kind = 4; break;
			case "PRODUCTIONS": t.kind = 5; break;
			case "END": t.kind = 6; break;
			case "WEAK": t.kind = 16; break;
			case "SYNC": t.kind = 17; break;
			case "ANY": t.kind = 18; break;
			case "IF": t.kind = 19; break;
			default: break;
		}
	}

	Token NextToken() {
		while (ignore[ch]) NextCh();

		t = new Token();
		t.pos = pos; t.col = pos - lineStart + 1; t.line = line; 
		int state = start[ch];
		tlen = 0; AddCh();
		
		switch (state) {
			case -1: { t.kind = eofSym; break; } // NextCh already done
			case 0: { t.kind = noSym; break; }   // NextCh already done
			case 1:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 1;}
				else {t.kind = 1; t.val = new String(tval, 0, tlen); CheckLiteral(); return t;}
			case 2:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 2;}
				else if (ch == '"') {AddCh(); goto case 4;}
				else if (ch == 92) {AddCh(); goto case 3;}
				else {t.kind = noSym; break;}
			case 3:
				if ((ch >= ' ' && ch <= '~')) {AddCh(); goto case 2;}
				else {t.kind = noSym; break;}
			case 4:
				{t.kind = 2; break;}
			case 5:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '&' || ch >= '(' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 6;}
				else if (ch == 92) {AddCh(); goto case 7;}
				else {t.kind = noSym; break;}
			case 6:
				if (ch == 39) {AddCh(); goto case 9;}
				else {t.kind = noSym; break;}
			case 7:
				if ((ch >= ' ' && ch <= '~')) {AddCh(); goto case 8;}
				else {t.kind = noSym; break;}
			case 8:
				if ((ch >= '0' && ch <= '9' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 8;}
				else if (ch == 39) {AddCh(); goto case 9;}
				else {t.kind = noSym; break;}
			case 9:
				{t.kind = 3; break;}
			case 10:
				if (!(ch == '*') && ch != Buffer.EOF) {AddCh(); goto case 10;}
				else if (ch == '*') {AddCh(); goto case 18;}
				else {t.kind = noSym; break;}
			case 11:
				{t.kind = 25; break;}
			case 12:
				if (!(ch == '*') && ch != Buffer.EOF) {AddCh(); goto case 12;}
				else if (ch == '*') {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 13:
				{t.kind = 26; break;}
			case 14:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 15;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 14;}
				else {t.kind = noSym; break;}
			case 15:
				{t.kind = 27; break;}
			case 16:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 16;}
				else {t.kind = 28; break;}
			case 17:
				if (ch == '*') {AddCh(); goto case 10;}
				else if (ch == '/') {AddCh(); goto case 14;}
				else {t.kind = noSym; break;}
			case 18:
				if ((ch <= ')' || ch >= '+' && ch <= '.' || ch >= '0' && ch <= 255)) {AddCh(); goto case 10;}
				else if (ch == '/') {AddCh(); goto case 11;}
				else if (ch == '*') {AddCh(); goto case 18;}
				else {t.kind = noSym; break;}
			case 19:
				if ((ch <= '(' || ch >= '+' && ch <= 255)) {AddCh(); goto case 12;}
				else if (ch == ')') {AddCh(); goto case 13;}
				else if (ch == '*') {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 20:
				{t.kind = 8; break;}
			case 21:
				{t.kind = 9; break;}
			case 22:
				{t.kind = 11; break;}
			case 23:
				{t.kind = 12; break;}
			case 24:
				{t.kind = 13; break;}
			case 25:
				{t.kind = 14; break;}
			case 26:
				{t.kind = 15; break;}
			case 27:
				{t.kind = 20; break;}
			case 28:
				{t.kind = 21; break;}
			case 29:
				{t.kind = 22; break;}
			case 30:
				{t.kind = 23; break;}
			case 31:
				if (ch == '*') {AddCh(); goto case 12;}
				else if (ch == '.') {AddCh(); goto case 29;}
				else {t.kind = 10; break;}
			case 32:
				if (ch == ')') {AddCh(); goto case 30;}
				else {t.kind = 7; break;}

		}
		t.val = new String(tval, 0, tlen);
		return t;
	}
	
	// get the next token (possibly a token already seen during peeking)
	public Token Scan () {
		if (tokens.next == null) {
			return NextToken();
		} else {
			pt = tokens = tokens.next;
			return tokens;
		}
	}

	// peek for the next token, ignore pragmas
	public Token Peek () {
		if (pt.next == null) {
			do {
				pt = pt.next = NextToken();
			} while (pt.kind > maxT); // skip pragmas
		} else {
			do {
				pt = pt.next;
			} while (pt.kind > maxT);
		}
		return pt;
	}
	
	// make sure that peeking starts at the current scan position
	public void ResetPeek () { pt = tokens; }

} // end Scanner

}