
using System;
using System.IO;
using System.Collections;

namespace Printy.Tool {

public class Token {
	public int kind;    // token kind
	public int pos;     // token position in the source text (starting at 0)
	public int col;     // token column (starting at 0)
	public int line;    // token line (starting at 1)
	public string val;  // token value
	public Token next;  // ML 2005-03-11 Tokens are kept in linked list
}

public class Buffer {
	public const char EOF = (char)256;
	const int MAX_BUFFER_LENGTH = 64 * 1024; // 64KB
	byte[] buf;         // input buffer
	int bufStart;       // position of first byte in buffer relative to input stream
	int bufLen;         // length of buffer
	int fileLen;        // length of input stream
	int pos;            // current position in buffer
	Stream stream;      // input stream (seekable)
	bool isUserStream;  // was the stream opened by the user?
	
	public Buffer (Stream s, bool isUserStream) {
		stream = s; this.isUserStream = isUserStream;
		fileLen = bufLen = (int) s.Length;
		if (stream.CanSeek && bufLen > MAX_BUFFER_LENGTH) bufLen = MAX_BUFFER_LENGTH;
		buf = new byte[bufLen];
		bufStart = Int32.MaxValue; // nothing in the buffer so far
		Pos = 0; // setup buffer to position 0 (start)
		if (bufLen == fileLen) Close();
	}
	
	~Buffer() { Close(); }
	
	void Close() {
		if (!isUserStream && stream != null) {
			stream.Close();
			stream = null;
		}
	}
	
	public int Read () {
		if (pos < bufLen) {
			return buf[pos++];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos++];
		} else {
			return EOF;
		}
	}

	public int Peek () {
		if (pos < bufLen) {
			return buf[pos];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos];
		} else {
			return EOF;
		}
	}
	
	public string GetString (int beg, int end) {
		int len = end - beg;
		char[] buf = new char[len];
		int oldPos = Pos;
		Pos = beg;
		for (int i = 0; i < len; i++) buf[i] = (char) Read();
		Pos = oldPos;
		return new String(buf);
	}

	public int Pos {
		get { return pos + bufStart; }
		set {
			if (value < 0) value = 0; 
			else if (value > fileLen) value = fileLen;
			if (value >= bufStart && value < bufStart + bufLen) { // already in buffer
				pos = value - bufStart;
			} else if (stream != null) { // must be swapped in
				stream.Seek(value, SeekOrigin.Begin);
				bufLen = stream.Read(buf, 0, buf.Length);
				bufStart = value; pos = 0;
			} else {
				pos = fileLen - bufStart; // make Pos return fileLen
			}
		}
	}
}

public class Scanner {
	const char EOL = '\n';
	const int eofSym = 0; /* pdt */
	const int charSetSize = 256;
	const int maxT = 36;
	const int noSym = 36;
	short[] start = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
	  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  0,  3,170,  0,171,  0,
	  0,175,  0, 83,101,151,  0,  0,  0,176,  0,  0,179,  0,  0,180,
	177,  0,140,178,  0,  0,  0,  0,  0,  0,  0, 54,  0, 55,  0,  0,
	  0, 20,  8,174, 13,  0,  0,  0, 37,  0,  0,  0,  0,  0, 26,173,
	  0,  0,  0,172,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  -1};


	public Buffer buffer; // scanner buffer
	
	Token t;          // current token
	char ch;          // current input character
	int pos;          // column number of current character
	int line;         // line number of current character
	int lineStart;    // start position of current line
	int oldEols;      // EOLs that appeared in a comment;
	BitArray ignore;  // set of characters to be ignored by the scanner

	Token tokens;     // list of tokens already peeked (first token is a dummy)
	Token pt;         // current peek token
	
	char[] tval = new char[128]; // text of current token
	int tlen;         // length of current token
	
	public Scanner (string fileName) {
		try {
			Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
			buffer = new Buffer(stream, false);
			Init();
		} catch (IOException) {
			Console.WriteLine("--- Cannot open file {0}", fileName);
			System.Environment.Exit(1);
		}
	}
	
	public Scanner (Stream s) {
		buffer = new Buffer(s, true);
		Init();
	}
	
	void Init() {
		pos = -1; line = 1; lineStart = 0;
		oldEols = 0;
		NextCh();
		ignore = new BitArray(charSetSize+1);
		ignore[' '] = true;  // blanks are always white space
		ignore[9] = true; ignore[10] = true; ignore[13] = true; 
		pt = tokens = new Token();  // first token is a dummy
	}
	
	void NextCh() {
		if (oldEols > 0) { ch = EOL; oldEols--; } 
		else {
			ch = (char)buffer.Read(); pos++;
			// replace isolated '\r' by '\n' in order to make
			// eol handling uniform across Windows, Unix and Mac
			if (ch == '\r' && buffer.Peek() != '\n') ch = EOL;
			if (ch == EOL) { line++; lineStart = pos + 1; }
		}

	}

	void AddCh() {
		if (tlen >= tval.Length) {
			char[] newBuf = new char[2 * tval.Length];
			Array.Copy(tval, 0, newBuf, 0, tval.Length);
			tval = newBuf;
		}
		tval[tlen++] = ch;
		NextCh();
	}



	bool Comment0() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '/') {
			NextCh();
			for(;;) {
				if (ch == 13) {
					NextCh();
					if (ch == 10) {
						level--;
						if (level == 0) { oldEols = line - line0; NextCh(); return true; }
						NextCh();
					}
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}

	bool Comment1() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '*') {
			NextCh();
			for(;;) {
				if (ch == '*') {
					NextCh();
					if (ch == '/') {
						level--;
						if (level == 0) { oldEols = line - line0; NextCh(); return true; }
						NextCh();
					}
				} else if (ch == '/') {
					NextCh();
					if (ch == '*') {
						level++; NextCh();
					}
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}


	void CheckLiteral() {
		switch (t.val) {
			default: break;
		}
	}

	Token NextToken() {
		while (ignore[ch]) NextCh();
		if (ch == '/' && Comment0() ||ch == '/' && Comment1()) return NextToken();
		t = new Token();
		t.pos = pos; t.col = pos - lineStart + 1; t.line = line; 
		int state = start[ch];
		tlen = 0; AddCh();
		
		switch (state) {
			case -1: { t.kind = eofSym; break; } // NextCh already done
			case 0: { t.kind = noSym; break; }   // NextCh already done
			case 1:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 2;}
				else {t.kind = noSym; break;}
			case 2:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 2;}
				else {t.kind = 1; break;}
			case 3:
				{t.kind = 2; break;}
			case 4:
				if (ch == 'a') {AddCh(); goto case 5;}
				else {t.kind = noSym; break;}
			case 5:
				if (ch == 'c') {AddCh(); goto case 6;}
				else {t.kind = noSym; break;}
			case 6:
				if (ch == 'e') {AddCh(); goto case 7;}
				else {t.kind = noSym; break;}
			case 7:
				{t.kind = 3; break;}
			case 8:
				if (ch == 'r') {AddCh(); goto case 9;}
				else {t.kind = noSym; break;}
			case 9:
				if (ch == 'e') {AddCh(); goto case 10;}
				else {t.kind = noSym; break;}
			case 10:
				if (ch == 'a') {AddCh(); goto case 11;}
				else {t.kind = noSym; break;}
			case 11:
				if (ch == 'k') {AddCh(); goto case 12;}
				else {t.kind = noSym; break;}
			case 12:
				{t.kind = 4; break;}
			case 13:
				if (ch == 'e') {AddCh(); goto case 14;}
				else {t.kind = noSym; break;}
			case 14:
				if (ch == 'f') {AddCh(); goto case 15;}
				else {t.kind = noSym; break;}
			case 15:
				if (ch == 'a') {AddCh(); goto case 16;}
				else {t.kind = noSym; break;}
			case 16:
				if (ch == 'u') {AddCh(); goto case 17;}
				else {t.kind = noSym; break;}
			case 17:
				if (ch == 'l') {AddCh(); goto case 18;}
				else {t.kind = noSym; break;}
			case 18:
				if (ch == 't') {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 19:
				{t.kind = 5; break;}
			case 20:
				if (ch == 'l') {AddCh(); goto case 21;}
				else {t.kind = noSym; break;}
			case 21:
				if (ch == 'w') {AddCh(); goto case 22;}
				else {t.kind = noSym; break;}
			case 22:
				if (ch == 'a') {AddCh(); goto case 23;}
				else {t.kind = noSym; break;}
			case 23:
				if (ch == 'y') {AddCh(); goto case 24;}
				else {t.kind = noSym; break;}
			case 24:
				if (ch == 's') {AddCh(); goto case 25;}
				else {t.kind = noSym; break;}
			case 25:
				{t.kind = 6; break;}
			case 26:
				if (ch == 'e') {AddCh(); goto case 27;}
				else {t.kind = noSym; break;}
			case 27:
				if (ch == 'v') {AddCh(); goto case 28;}
				else {t.kind = noSym; break;}
			case 28:
				if (ch == 'e') {AddCh(); goto case 29;}
				else {t.kind = noSym; break;}
			case 29:
				if (ch == 'r') {AddCh(); goto case 30;}
				else {t.kind = noSym; break;}
			case 30:
				{t.kind = 7; break;}
			case 31:
				if (ch == 'e') {AddCh(); goto case 32;}
				else {t.kind = noSym; break;}
			case 32:
				if (ch == 'n') {AddCh(); goto case 33;}
				else {t.kind = noSym; break;}
			case 33:
				{t.kind = 8; break;}
			case 34:
				if (ch == 's') {AddCh(); goto case 35;}
				else {t.kind = noSym; break;}
			case 35:
				if (ch == 'e') {AddCh(); goto case 36;}
				else {t.kind = noSym; break;}
			case 36:
				{t.kind = 9; break;}
			case 37:
				if (ch == 'i') {AddCh(); goto case 38;}
				else {t.kind = noSym; break;}
			case 38:
				if (ch == 'd') {AddCh(); goto case 39;}
				else {t.kind = noSym; break;}
			case 39:
				if (ch == 'e') {AddCh(); goto case 40;}
				else {t.kind = noSym; break;}
			case 40:
				{t.kind = 10; break;}
			case 41:
				if (ch == 'e') {AddCh(); goto case 42;}
				else {t.kind = noSym; break;}
			case 42:
				if (ch == 'r') {AddCh(); goto case 43;}
				else {t.kind = noSym; break;}
			case 43:
				if (ch == 'l') {AddCh(); goto case 44;}
				else {t.kind = noSym; break;}
			case 44:
				if (ch == 'a') {AddCh(); goto case 45;}
				else {t.kind = noSym; break;}
			case 45:
				if (ch == 'p') {AddCh(); goto case 46;}
				else {t.kind = noSym; break;}
			case 46:
				{t.kind = 11; break;}
			case 47:
				if (ch == 'y') {AddCh(); goto case 48;}
				else {t.kind = noSym; break;}
			case 48:
				if (ch == 'l') {AddCh(); goto case 49;}
				else {t.kind = noSym; break;}
			case 49:
				if (ch == 'e') {AddCh(); goto case 50;}
				else {t.kind = noSym; break;}
			case 50:
				{t.kind = 12; break;}
			case 51:
				if (ch == 's') {AddCh(); goto case 52;}
				else {t.kind = noSym; break;}
			case 52:
				if (ch == 's') {AddCh(); goto case 53;}
				else {t.kind = noSym; break;}
			case 53:
				{t.kind = 13; break;}
			case 54:
				{t.kind = 14; break;}
			case 55:
				{t.kind = 15; break;}
			case 56:
				if (ch == 'L') {AddCh(); goto case 57;}
				else {t.kind = noSym; break;}
			case 57:
				{t.kind = 16; break;}
			case 58:
				if (ch == 'D') {AddCh(); goto case 59;}
				else {t.kind = noSym; break;}
			case 59:
				if (ch == 'E') {AddCh(); goto case 60;}
				else {t.kind = noSym; break;}
			case 60:
				if (ch == 'N') {AddCh(); goto case 61;}
				else {t.kind = noSym; break;}
			case 61:
				if (ch == 'T') {AddCh(); goto case 62;}
				else {t.kind = noSym; break;}
			case 62:
				{t.kind = 17; break;}
			case 63:
				if (ch == 'u') {AddCh(); goto case 64;}
				else {t.kind = noSym; break;}
			case 64:
				if (ch == 'c') {AddCh(); goto case 65;}
				else {t.kind = noSym; break;}
			case 65:
				if (ch == 't') {AddCh(); goto case 66;}
				else {t.kind = noSym; break;}
			case 66:
				if (ch == 'i') {AddCh(); goto case 67;}
				else {t.kind = noSym; break;}
			case 67:
				if (ch == 'o') {AddCh(); goto case 68;}
				else {t.kind = noSym; break;}
			case 68:
				if (ch == 'n') {AddCh(); goto case 69;}
				else {t.kind = noSym; break;}
			case 69:
				{t.kind = 18; break;}
			case 70:
				if (ch == 'm') {AddCh(); goto case 71;}
				else {t.kind = noSym; break;}
			case 71:
				if (ch == 'b') {AddCh(); goto case 72;}
				else {t.kind = noSym; break;}
			case 72:
				if (ch == 'o') {AddCh(); goto case 73;}
				else {t.kind = noSym; break;}
			case 73:
				if (ch == 'l') {AddCh(); goto case 74;}
				else {t.kind = noSym; break;}
			case 74:
				{t.kind = 19; break;}
			case 75:
				if (ch == 'e') {AddCh(); goto case 76;}
				else {t.kind = noSym; break;}
			case 76:
				if (ch == 'r') {AddCh(); goto case 77;}
				else {t.kind = noSym; break;}
			case 77:
				if (ch == 'a') {AddCh(); goto case 78;}
				else {t.kind = noSym; break;}
			case 78:
				if (ch == 'l') {AddCh(); goto case 79;}
				else {t.kind = noSym; break;}
			case 79:
				{t.kind = 20; break;}
			case 80:
				if (ch == 'e') {AddCh(); goto case 81;}
				else {t.kind = noSym; break;}
			case 81:
				if (ch == 'n') {AddCh(); goto case 82;}
				else {t.kind = noSym; break;}
			case 82:
				{t.kind = 21; break;}
			case 83:
				if (ch == 'l') {AddCh(); goto case 84;}
				else {t.kind = noSym; break;}
			case 84:
				if (ch == 'o') {AddCh(); goto case 85;}
				else {t.kind = noSym; break;}
			case 85:
				if (ch == 's') {AddCh(); goto case 86;}
				else {t.kind = noSym; break;}
			case 86:
				if (ch == 'e') {AddCh(); goto case 87;}
				else {t.kind = noSym; break;}
			case 87:
				{t.kind = 22; break;}
			case 88:
				if (ch == 's') {AddCh(); goto case 89;}
				else {t.kind = noSym; break;}
			case 89:
				if (ch == 'i') {AddCh(); goto case 90;}
				else {t.kind = noSym; break;}
			case 90:
				if (ch == 'g') {AddCh(); goto case 91;}
				else {t.kind = noSym; break;}
			case 91:
				if (ch == 'n') {AddCh(); goto case 92;}
				else {t.kind = noSym; break;}
			case 92:
				{t.kind = 23; break;}
			case 93:
				if (ch == 'p') {AddCh(); goto case 94;}
				else {t.kind = noSym; break;}
			case 94:
				if (ch == 'a') {AddCh(); goto case 95;}
				else {t.kind = noSym; break;}
			case 95:
				if (ch == 'r') {AddCh(); goto case 96;}
				else {t.kind = noSym; break;}
			case 96:
				if (ch == 'a') {AddCh(); goto case 97;}
				else {t.kind = noSym; break;}
			case 97:
				if (ch == 't') {AddCh(); goto case 98;}
				else {t.kind = noSym; break;}
			case 98:
				if (ch == 'o') {AddCh(); goto case 99;}
				else {t.kind = noSym; break;}
			case 99:
				if (ch == 'r') {AddCh(); goto case 100;}
				else {t.kind = noSym; break;}
			case 100:
				{t.kind = 24; break;}
			case 101:
				if (ch == 'e') {AddCh(); goto case 102;}
				else {t.kind = noSym; break;}
			case 102:
				if (ch == 'l') {AddCh(); goto case 103;}
				else {t.kind = noSym; break;}
			case 103:
				if (ch == 'i') {AddCh(); goto case 104;}
				else {t.kind = noSym; break;}
			case 104:
				if (ch == 'm') {AddCh(); goto case 105;}
				else {t.kind = noSym; break;}
			case 105:
				if (ch == 'i') {AddCh(); goto case 106;}
				else {t.kind = noSym; break;}
			case 106:
				if (ch == 't') {AddCh(); goto case 107;}
				else {t.kind = noSym; break;}
			case 107:
				if (ch == 'e') {AddCh(); goto case 108;}
				else {t.kind = noSym; break;}
			case 108:
				if (ch == 'r') {AddCh(); goto case 109;}
				else {t.kind = noSym; break;}
			case 109:
				{t.kind = 25; break;}
			case 110:
				if (ch == 'n') {AddCh(); goto case 111;}
				else {t.kind = noSym; break;}
			case 111:
				if (ch == 'e') {AddCh(); goto case 112;}
				else {t.kind = noSym; break;}
			case 112:
				if (ch == 'r') {AddCh(); goto case 113;}
				else {t.kind = noSym; break;}
			case 113:
				if (ch == 'C') {AddCh(); goto case 114;}
				else {t.kind = noSym; break;}
			case 114:
				if (ch == 'o') {AddCh(); goto case 115;}
				else {t.kind = noSym; break;}
			case 115:
				if (ch == 'm') {AddCh(); goto case 116;}
				else {t.kind = noSym; break;}
			case 116:
				if (ch == 'm') {AddCh(); goto case 117;}
				else {t.kind = noSym; break;}
			case 117:
				if (ch == 'e') {AddCh(); goto case 118;}
				else {t.kind = noSym; break;}
			case 118:
				if (ch == 'n') {AddCh(); goto case 119;}
				else {t.kind = noSym; break;}
			case 119:
				if (ch == 't') {AddCh(); goto case 120;}
				else {t.kind = noSym; break;}
			case 120:
				{t.kind = 26; break;}
			case 121:
				if (ch == 't') {AddCh(); goto case 122;}
				else {t.kind = noSym; break;}
			case 122:
				if (ch == 'e') {AddCh(); goto case 123;}
				else {t.kind = noSym; break;}
			case 123:
				if (ch == 'r') {AddCh(); goto case 124;}
				else {t.kind = noSym; break;}
			case 124:
				if (ch == 'C') {AddCh(); goto case 125;}
				else {t.kind = noSym; break;}
			case 125:
				if (ch == 'o') {AddCh(); goto case 126;}
				else {t.kind = noSym; break;}
			case 126:
				if (ch == 'm') {AddCh(); goto case 127;}
				else {t.kind = noSym; break;}
			case 127:
				if (ch == 'm') {AddCh(); goto case 128;}
				else {t.kind = noSym; break;}
			case 128:
				if (ch == 'e') {AddCh(); goto case 129;}
				else {t.kind = noSym; break;}
			case 129:
				if (ch == 'n') {AddCh(); goto case 130;}
				else {t.kind = noSym; break;}
			case 130:
				if (ch == 't') {AddCh(); goto case 131;}
				else {t.kind = noSym; break;}
			case 131:
				{t.kind = 27; break;}
			case 132:
				if (ch == 't') {AddCh(); goto case 133;}
				else {t.kind = noSym; break;}
			case 133:
				if (ch == 'r') {AddCh(); goto case 134;}
				else {t.kind = noSym; break;}
			case 134:
				if (ch == 'i') {AddCh(); goto case 135;}
				else {t.kind = noSym; break;}
			case 135:
				if (ch == 'b') {AddCh(); goto case 136;}
				else {t.kind = noSym; break;}
			case 136:
				if (ch == 'u') {AddCh(); goto case 137;}
				else {t.kind = noSym; break;}
			case 137:
				if (ch == 't') {AddCh(); goto case 138;}
				else {t.kind = noSym; break;}
			case 138:
				if (ch == 'e') {AddCh(); goto case 139;}
				else {t.kind = noSym; break;}
			case 139:
				{t.kind = 28; break;}
			case 140:
				if (ch == 'e') {AddCh(); goto case 141;}
				else {t.kind = noSym; break;}
			case 141:
				if (ch == 's') {AddCh(); goto case 142;}
				else {t.kind = noSym; break;}
			case 142:
				if (ch == 'o') {AddCh(); goto case 143;}
				else {t.kind = noSym; break;}
			case 143:
				if (ch == 'l') {AddCh(); goto case 144;}
				else {t.kind = noSym; break;}
			case 144:
				if (ch == 'v') {AddCh(); goto case 145;}
				else {t.kind = noSym; break;}
			case 145:
				if (ch == 'e') {AddCh(); goto case 146;}
				else {t.kind = noSym; break;}
			case 146:
				if (ch == 'r') {AddCh(); goto case 147;}
				else {t.kind = noSym; break;}
			case 147:
				{t.kind = 29; break;}
			case 148:
				if (ch == 'o') {AddCh(); goto case 149;}
				else {t.kind = noSym; break;}
			case 149:
				if (ch == 'g') {AddCh(); goto case 150;}
				else {t.kind = noSym; break;}
			case 150:
				{t.kind = 30; break;}
			case 151:
				if (ch == 'p') {AddCh(); goto case 152;}
				else {t.kind = noSym; break;}
			case 152:
				if (ch == 'i') {AddCh(); goto case 153;}
				else {t.kind = noSym; break;}
			case 153:
				if (ch == 'l') {AddCh(); goto case 154;}
				else {t.kind = noSym; break;}
			case 154:
				if (ch == 'o') {AddCh(); goto case 155;}
				else {t.kind = noSym; break;}
			case 155:
				if (ch == 'g') {AddCh(); goto case 156;}
				else {t.kind = noSym; break;}
			case 156:
				{t.kind = 31; break;}
			case 157:
				if (ch == 't') {AddCh(); goto case 158;}
				else {t.kind = noSym; break;}
			case 158:
				if (ch == 'i') {AddCh(); goto case 159;}
				else {t.kind = noSym; break;}
			case 159:
				if (ch == 'o') {AddCh(); goto case 160;}
				else {t.kind = noSym; break;}
			case 160:
				if (ch == 'n') {AddCh(); goto case 161;}
				else {t.kind = noSym; break;}
			case 161:
				{t.kind = 32; break;}
			case 162:
				if (ch == 'e') {AddCh(); goto case 163;}
				else {t.kind = noSym; break;}
			case 163:
				if (ch == 'N') {AddCh(); goto case 164;}
				else {t.kind = noSym; break;}
			case 164:
				if (ch == 'u') {AddCh(); goto case 165;}
				else {t.kind = noSym; break;}
			case 165:
				if (ch == 'm') {AddCh(); goto case 166;}
				else {t.kind = noSym; break;}
			case 166:
				if (ch == 'b') {AddCh(); goto case 167;}
				else {t.kind = noSym; break;}
			case 167:
				if (ch == 'e') {AddCh(); goto case 168;}
				else {t.kind = noSym; break;}
			case 168:
				if (ch == 'r') {AddCh(); goto case 169;}
				else {t.kind = noSym; break;}
			case 169:
				{t.kind = 33; break;}
			case 170:
				{t.kind = 34; break;}
			case 171:
				{t.kind = 35; break;}
			case 172:
				if (ch == 'p') {AddCh(); goto case 4;}
				else if (ch == 't') {AddCh(); goto case 47;}
				else {t.kind = noSym; break;}
			case 173:
				if (ch == 'p') {AddCh(); goto case 31;}
				else if (ch == 'v') {AddCh(); goto case 41;}
				else {t.kind = noSym; break;}
			case 174:
				if (ch == 'l') {AddCh(); goto case 181;}
				else {t.kind = noSym; break;}
			case 175:
				if (ch == 'L') {AddCh(); goto case 56;}
				else if (ch == 's') {AddCh(); goto case 88;}
				else if (ch == 't') {AddCh(); goto case 132;}
				else if (ch == 'c') {AddCh(); goto case 157;}
				else {t.kind = noSym; break;}
			case 176:
				if (ch == 'N') {AddCh(); goto case 58;}
				else if (ch == 'n') {AddCh(); goto case 110;}
				else {t.kind = noSym; break;}
			case 177:
				if (ch == 'r') {AddCh(); goto case 182;}
				else {t.kind = noSym; break;}
			case 178:
				if (ch == 'y') {AddCh(); goto case 70;}
				else if (ch == 'e') {AddCh(); goto case 93;}
				else {t.kind = noSym; break;}
			case 179:
				if (ch == 'i') {AddCh(); goto case 183;}
				else {t.kind = noSym; break;}
			case 180:
				if (ch == 'p') {AddCh(); goto case 80;}
				else if (ch == 'u') {AddCh(); goto case 121;}
				else {t.kind = noSym; break;}
			case 181:
				if (ch == 'o') {AddCh(); goto case 34;}
				else if (ch == 'a') {AddCh(); goto case 51;}
				else {t.kind = noSym; break;}
			case 182:
				if (ch == 'o') {AddCh(); goto case 184;}
				else {t.kind = noSym; break;}
			case 183:
				if (ch == 't') {AddCh(); goto case 75;}
				else if (ch == 'n') {AddCh(); goto case 162;}
				else {t.kind = noSym; break;}
			case 184:
				if (ch == 'd') {AddCh(); goto case 63;}
				else if (ch == 'l') {AddCh(); goto case 148;}
				else {t.kind = noSym; break;}

		}
		t.val = new String(tval, 0, tlen);
		return t;
	}
	
	// get the next token (possibly a token already seen during peeking)
	public Token Scan () {
		if (tokens.next == null) {
			return NextToken();
		} else {
			pt = tokens = tokens.next;
			return tokens;
		}
	}

	// peek for the next token, ignore pragmas
	public Token Peek () {
		if (pt.next == null) {
			do {
				pt = pt.next = NextToken();
			} while (pt.kind > maxT); // skip pragmas
		} else {
			do {
				pt = pt.next;
			} while (pt.kind > maxT);
		}
		return pt;
	}
	
	// make sure that peeking starts at the current scan position
	public void ResetPeek () { pt = tokens; }

} // end Scanner

}